////////////////////////////////////////////////////////////////////////////////
// c3000 project
//
// This is a video card project for the Carte Blanche II.
// It is for use with the Apple //e only. 
// It should run in any slot but has not been tested in them all.
//
// This project is a port of the c302 project for the Carte Blanche I with a
// couple of additions.
// 
// That version is originally based on the video output functions in
// Alex Freed's JAT (Jack of All Trades) project for the CB1.   
//
// The screen resolution is 640 x 480 @ 60 Hz.  The DCM (main clock) uses the 
// on board 12 MHz oscillator.  The pixel clock is 25 MHz (slightly lower than
// the standard). 
// 
// The video is output as both HDMI and VGA.  All common //e video modes are 
// supported and in addition there is the ability to set the text, background
// and border colors similar to the Apple IIgs.
// IIgs style Super Hires (SHR) Graphics modes are also supported.
//	The SHR 640 mode can be viewed in as dithered or not dithered. 
// A grayscale Double Hires mode is also available.
// The //e LORES, Double LORES and Double HIRES can be viewed with either one
// gray color (normal for the //e) or two grays like the IIgs.
// These extra modes and settings can be manipulated with softswitches and 
// there is also a mechanism for saving them to the CB2 external flash memory.
// On start up the bitstream auto-loads the saved settings.
// 
// This project also contains code to flash a bitstream to the external flash
// memory chip.
//
// The bitstream was generated using Xilinx ISE version 14.7.
//
// See the c3000_notes file for more information, bugs, limitations, etc.
// 
////////////////////////////////////////////////////////////////////////////////
// cb2.v (module name 'C3000')
// Top level module for the project.
// Contains code for:
// Instantiating sub modules.
// Reading and writing SRAM, BRAM, distributed RAM and arrays of registers.
// Synchronizing Apple signals to CB clock.
// Slot device I/O ($C0nx).
// DMA & interrupt priority daisy chain.
// Reading / Writing to the CB2 external flash. 
//
////////////////////////////////////////////////////////////////////////////////
`include "c3000_defs.v"
////////////////////////////////////////////////////////////////////////////////
module C3000 		// this version is for the //e  
  (					
////////////////////////////////////////////////////////////////////////////////
// inputs/outputs to module						
   input	 	BUS_7M,			// 7 MHz Apple clock
   inout [15:0]  BUS_A,		// Apple address BUS
   inout [7:0]  BUS_D,		// Apple data BUS
   input 	BUS_DMAIN,		// DMA priority daisy chain
   output   BUS_DMAOUT,		// DMA priority daisy chain
   input 	BUS_INTIN,		// interrupt priority daisy chain
   output   BUS_INTOUT,		// interrupt priority daisy chain
   input    BUS_PHASE0,		// phase 0 clock
   input    BUS_PHASE1,		// phase 1 clock
   input    BUS_Q3,			// Q3 clock
   input    BUS_RW,			// read/write signal
   input 	BUS_RESET,		// reset
   input    BUS_DEV_SEL,	// Apple slot device select
	output	ABUS_EN,
	output 	USER_LED,		// blue LED 0 = on
////////////////////////////////////////////////////////////////////////////////
// hdmi/dvi & vga 
	output [7:0] RED,			// 8 bit red to hdmi	and vga
	output [7:0] GREEN,		// 8 bit green to hdmi and vga
	output [7:0] BLUE,		// 8 bit blue to hdmi and vga
	output HSYNC,				// horizontal sync pulse to hdmi & VGA	
	output VSYNC,				// vertical sync pulse to hdmi & VGA
	output FRAME_ENABLE,		// frame enable signal to hdmi 	
	output PIX_CLK, 			// VGA & hdmi pixel clock			
	output LCD_EN,
	output DAC_EN,				// VGA on 
	output DAC_SONG,			// Sync on Green (grounded)
	input MSEN,					// monitor detect  0 = detected (HDMI) only
   output [17:0]   SRAM_A,	// SRAM address					
   inout  [7:0]    SRAM_D,	// SRAM data							
   output   SRAM_NCS,		// SRAM enable chip 0 (low)		
   output   SRAM_NOE,		// SRAM output enable (low)
   output   SRAM_NWE,		// SRAM write enable	(low)	
	output 	SRAM_NBE0,		// SRAM low data byte (low)
	output 	SRAM_NBE1,		// SRAM high data byte (low)
   input 	SYSCLK,			// 12 MHz onboard CB clock	
   input    FPGA_MISO, 		// data from device
   output   FPGA_MOSI, 		// data to device
   output   FPGA_SPICLK, 	// devide clock
   output reg FLASH_CS		// select external flash memory
  );
////////////////////////////////////////////////////////////////////////////////
// version 
	reg [7:0] maj_version = 30;  // these can be viewed with a slot IO command
	reg [7:0] min_version = 00;
////////////////////////////////////////////////////////////////////////////////
// Apple clocks sync'd to clk50 
	reg reset_s;	 
	reg phase0_s;	
	reg Q3_s;		
	reg devsel_s;		
	reg iosel_s;	
	reg iostrobe_s;					
	reg[15:0] bus_a_1;	
	reg[15:0] bus_a_s; 	
	reg[7:0] bus_d_1;		
	reg[7:0] bus_d_s;
	reg ph1_s;		
////////////////////////////////////////////////////////////////////////////////
// debug - increments border color
`ifdef bbug
	wire bugin;							// used for debugging
`endif 
////////////////////////////////////////////////////////////////////////////////
// Apple bus on
	assign ABUS_EN = 1'b0;			// no Apple bus IO if this is 1'b1
////////////////////////////////////////////////////////////////////////////////
// video connections
	wire h_sync;						// h sync from vga.video.v
	wire v_sync;						// v sync from vga.video.v
	wire frame_en;						// when pixels are sent to screen (from vga.video.v)
	assign HSYNC = h_sync;			// h sync to hdmi & VGA chips
	assign VSYNC = v_sync;			// v sync to hdmi & VGA chips
	assign FRAME_ENABLE = frame_en;	// when to draw pixels 
	assign LCD_EN = frame_en;		
	assign DAC_EN = frame_en;		// turn on VGA output when sending pixels
	assign DAC_SONG = 1'b0;			// grounded for VGA			
////////////////////////////////////////////////////////////////////////////////
// user LED
`ifdef bbug
	assign USER_LED = ~bugin; 				// 1 = off - 0 is blue LED on
`else
	assign USER_LED = 1'b1;					// off
`endif
////////////////////////////////////////////////////////////////////////////////
//	clock 
	reg clk25_b;		// 25 MHz used with the SPI functions
	wire clk25;			// 25 MHz derived from clk50 
   wire CLKFX_OUT;	// 50 MHz main clock
   wire LOCKED_OUT;	
   wire clk50 = CLKFX_OUT;
// instantiate DCM for main clock
   clk0 clock (
	       .CLKIN_IN(SYSCLK),			// input 12 MHz CB2 clock  
	       .RST_IN(1'b0),				
	       .CLKFX_OUT(CLKFX_OUT), 	// output 50 MHz main clock
	       .LOCKED_OUT(LOCKED_OUT)	// 1 = clock is locked (operating correctly)
	       );
////////////////////////////////////////////////////////////////////////////////
// IO - used to pass new softswitch settings from vga_video.v 	
	wire io_read;
	wire [7:0] io_reg;
////////////////////////////////////////////////////////////////////////////////	
//   connect daisy chain pins
   assign 	BUS_DMAOUT = BUS_DMAIN;		// just connect the DMA chain 
	assign 	BUS_INTOUT = BUS_INTIN;		// just connect the interrupt chain
////////////////////////////////////////////////////////////////////////////////
// write_clk signals	for buffering the Apple II video data
   wire video_we;						// RAM write enable (from wrclk.v) 
   wire video_write_SRAM;			// OK to write Apple video data to SRAM
											// or BRAM (from wrclk.v) 
   wire [15:0] video_wr_addr;		// address of apple video data  
   wire [7:0] 	write_data;			// apple video data

	wire update_time;					// OK to update the CB buffer	(from vga_video.v)
											// passed to wrclk.v
////////////////////////////////////////////////////////////////////////////////
// instantiate write_clk (wrclk.v) module
   write_clk 
     write_clk (
		.clk(clk50), 									// master clock from DCM
		.ph0(BUS_PHASE0), 							// Apple phase 0 clock
		.Q3(BUS_Q3), 									// Q3 Apple clock
		.rw(BUS_RW),									// read/write signal from Apple
		.update_time(update_time), 				// when it is OK to get Apple data 
		.bus_a_s(bus_a_s), 							// Apple memory address
		.bus_d_s(bus_d_s),							// Apple memory data
		.we(video_we),									// OK to write enable CB video buffers
		.video_write_SRAM(video_write_SRAM),	// Ok to write to CB video buffers
		.ram_addr(video_wr_addr), 					// address of data to buffer 
		.write_data(write_data)						// data to buffer
		);
////////////////////////////////////////////////////////////////////////////////	
// video buffer access wires
   wire [10:0] rom_addr;			// address of character data from vga_video.v
   wire [7:0] 	char_data; 			// character sent to vga_video.v 
   wire [7:0] 	video_ram_data;	// data read from CB video buffers
   wire [16:0] ram_address;		// address used to read from video buffer  
   wire 	auxmem; 						// 1 means the apple wrote a bank 1 video
											// address (aux memory) - calculate in vga_video.v
////////////////////////////////////////////////////////////////////////////////
// Main memory video data from $2000 to $5FFF is buffered in FPGA BlockRAM (BRAM)
////////////////////////////////////////////////////////////////////////////////
// write_vid_is_hgr 
// high if Apple is writing to video address $2000 to $5FFFF in main memory, 
// not aux memory
//	video_wr_addr (16 bit) is the ram address output from the write_clk module 
// on the falling edge of Q3 and rw == 0
	wire write_vid_is_hgr = (video_wr_addr > 16'h1FFF) & (video_wr_addr < 16'h6000) & ~auxmem;
// read_vid_is_hgr
// high if buffer read address (from vga_video.v) is $2000 to $5FFF in main memory, not aux memory
// ram_address is the address from the vga_video module to read data from the
// SRAM or BRAM (on FPGA block RAM)
	wire read_vid_is_hgr = (ram_address[15:0] > 16'h1FFF) & (ram_address[15:0] < 16'h6000) & (ram_address[16] == 1'b0);
// Block RAM buffer I/O for main memory video data $2000 to $5FFFF  
	wire[13:0] BRAM_AA_hgr;
	wire[13:0] BRAM_AB_hgr;
	wire[7:0] BRAM_DA_hgr;
	wire[7:0] BRAM_DB_hgr;
// instantiate - hgr_buf - simple dual port RAM
	hgr_buf hgr_buf(			// buffers main RAM HGR & DHR both pages
	.clka(clk50),				// clock for writing
	.wea(video_we),  			// enable writing
	.addra(BRAM_AA_hgr), 	// write address 14 bit
	.dina(BRAM_DA_hgr), 		// write data 
	.clkb(clk50),				// clock for reading
	.addrb(BRAM_AB_hgr), 	// read address 14 bit 
	.doutb(BRAM_DB_hgr)); 	// read data 
// hgr buffer address & data logic
	assign 	BRAM_AA_hgr = (video_write_SRAM && write_vid_is_hgr) ? video_wr_addr[13:0] : 14'bz; 	// write
	assign 	BRAM_AB_hgr = (read_vid_is_hgr) ? ram_address[13:0] : 14'bz;									// read addr	
	assign	BRAM_DA_hgr = (video_write_SRAM && write_vid_is_hgr) ? write_data : 8'bz;
////////////////////////////////////////////////////////////////////////////////
// Text memory data from $400 to $0BFF main & aux is buffered in FPGA 
// BlockRAM (BRAM) - text_buf
// 
// write_vid_is_text
// high if Apple is writing address $0400 to $0BFF (TEXT pg 1 or 2 in either bank 
	wire write_vid_is_text = (video_wr_addr > 16'h03FF) & (video_wr_addr < 16'h0C00);
// read_vid_is_text
// high if buffer read address (from vga_video.v) is $0400 to $0BFF in either bank 0 or 1
	wire read_vid_is_text = (ram_address[15:0] > 16'h03FF) & (ram_address[15:0] < 16'h0C00);
// Block RAM buffer I/O for text memory 
	wire[11:0] BRAM_AA;	// write addr
	wire[11:0] BRAM_AB;	// read addr
	wire[7:0] BRAM_DA;	// write data
	wire[7:0] BRAM_DB;	// read data
// instantiate - text_buf - simple dual port RAM
	text_buf text_buf (	// buffers text pages 1 & 2, main and aux
		.clka(clk50),			// clock for writing
		.wea(video_we),   	// enable writing
		.addra(BRAM_AA), 		// write address 12 bit
		.dina(BRAM_DA),  		// write data
		.clkb(clk50),			// clock for reading
		.addrb(BRAM_AB), 		// read address 12 bit 
		.doutb(BRAM_DB)); 	// read data
// TEXT buffer address & data logic 
	assign 	BRAM_AA = (video_write_SRAM && write_vid_is_text) ? {auxmem, video_wr_addr[10:0]} : 12'bz; 	// write addr
	assign 	BRAM_AB = (read_vid_is_text) ? {ram_address[16], ram_address[10:0]} : 12'bz;						// read addr
	assign	BRAM_DA = (video_write_SRAM && write_vid_is_text) ? write_data : 8'bz;
////////////////////////////////////////////////////////////////////////////////	
// Palettes (in SHR) $019E00 to $019FFF are buffered in FPGA 
// BlockRAM (BRAM) - pal_buf
//
// write_is_pal
// high if Apple is writing address $019E00 to $019FFF
	wire write_is_pal =  ((video_wr_addr > 16'h9DFF) && (video_wr_addr < 16'hA000) && auxmem ) ? 1'b1 : 1'b0;
// write_is_even 
// high if the byte written is even	
	wire write_is_even = (video_wr_addr[0] == 1'b0) ? 1'b1 : 1'b0; 
// Block RAM buffer I/O for SHR palettes
	wire[7:0] BRAM_AA_PAL;		// write addr
	wire[7:0] BRAM_AB_PAL;		// read addr
	wire[11:0] BRAM_DA_PAL;		// write data
	wire[11:0] BRAM_DB_PAL;		// read data
// instantiate - pal_buf - simple dual port RAM
	pal_buf pal_buf (				// buffers palettes 12 bits
										// not used if pal_fixed defined in video module
		.clka(clk50),				// write clock
		.wea(video_we), 			// enable writing
		.addra(BRAM_AA_PAL), 	// write address 8 bit
		.dina(BRAM_DA_PAL), 		// write data 12 bit
		.clkb(clk50),				// read clock
		.addrb(BRAM_AB_PAL), 	// read address 8 bit 
		.doutb(BRAM_DB_PAL)); 	// read data 12 bit 
// use palette address to get palette data
	wire[7:0] pal_addr; 	// read address from vga_video.v to BRAM - pal_buf
	wire[11:0] pal_data; // read data to vga_video.v from BRAM - pal_buf	
// odd-even palette registers
// palettes are 12-bits store in 2 bytes (16 bits) 4 bits in odd bytes and
// 8 bits in even bytes
	reg [7:0] even_pal = 0;	// 8 bits		
	reg [3:0] odd_pal = 0;	// 4 bits

// palette address written to BRAM - pal_buf
	assign	BRAM_AA_PAL = (video_write_SRAM && write_is_pal) ? (video_wr_addr[8:1]) : 8'bz;	
// palette data written to BRAM - pal_buf
	assign	BRAM_DA_PAL = (video_write_SRAM && write_is_pal) ? {odd_pal, even_pal} : 12'bz;
// palette address for reading palette data from BRAM - pal_buf
	assign	BRAM_AB_PAL = pal_addr;
// palette data read from BRAM - pal_buf
	assign 	pal_data = BRAM_DB_PAL;

// video_mem
// high when HGR, DHR or SHR data is written by Apple	
	wire video_mem = ((video_wr_addr > 16'h1FFF) && (video_wr_addr < 16'h9D00)) ? 1'b1 : 1'b0;	
// vid_read_mem
// high when HGR, DHR or SHR data is being read by VGA/HDMI (vga_video.v) 
	wire vid_read_mem = ((ram_address[15:0] > 16'h1FFF) && (ram_address[15:0] < 16'h9D00)) ? 1'b1 : 1'b0;
// write_vid_is_gs
// high when SHR palettes or scan control bytes are being written by Apple
	wire write_vid_is_gs = ((video_wr_addr > 16'h9CFF) && (video_wr_addr < 16'hA000) && auxmem) ? 1'b1 : 1'b0;
// flashing takes a free ride on the video code
	wire write_vid_is_ff = ((video_wr_addr > 16'h3FFF) && (video_wr_addr < 16'h4100) && ~auxmem && video_write_SRAM);
// high when Scanline Control bytes are written by the Apple II
	wire write_is_scb = ((video_wr_addr > 16'h9CFF) && (video_wr_addr < 16'h9DC8) && auxmem) ? 1'b1 : 1'b0;
////////////////////////////////////////////////////////////////////////////////
// video data for output from both SRAM & BRAM	
	assign 	video_ram_data = (~video_write_SRAM & vid_read_mem & ~read_vid_is_hgr) ?  SRAM_D : read_vid_is_text ? BRAM_DB : (~video_write_SRAM & read_vid_is_hgr) ? BRAM_DB_hgr : 8'bz;	
////////////////////////////////////////////////////////////////////////////////
// SRAM address and data
	assign 	SRAM_A = (video_write_SRAM && ~write_vid_is_text && ~write_vid_is_hgr  && ~write_vid_is_gs) ? {auxmem, video_wr_addr} : (vid_read_mem  && ~read_vid_is_hgr) ? ram_address : 17'bz;
	assign 	SRAM_D = (video_write_SRAM & video_mem & ~write_vid_is_hgr && ~write_vid_is_gs) ? write_data : 8'bz;	
////////////////////////////////////////////////////////////////////////////////
// SRAM chip select	
   assign 	SRAM_NCS = 1'b0;			// use SRAM chip 0
	assign	SRAM_NBE0 = 1'b0;			// low byte
	assign 	SRAM_NBE1 = 1'b1;			// high byte 
////////////////////////////////////////////////////////////////////////////////
// SRAM chip read / write 	
	assign 	SRAM_NOE = 1'b0; 			// SRAM datasheet says /OE can be low for both read and write
   assign 	SRAM_NWE = ~video_we;
////////////////////////////////////////////////////////////////////////////////
// scan control bytes

	wire[7:0] scb_addr; 				// read address
	wire[7:0] scb_data; 				// read data
	reg [7:0] scb [199:0];			// 200 - 8 bit registers for scan control bytes

	reg [7:0] ff_array [255:0];	// flashing takes a free ride on the video 
											// routines 
	
	reg [255:0] k;						// 256 - 8 bit registers for cb2flash data

// initialize the scanline control byte registers & the flash data registers
	initial
		begin
		for (k = 0; k < 200; k = k + 1)	// 200 scanline registers
			begin
			scb[k] = 0;
			end

		for (k = 0; k < 256; k = k + 1)	// 256 registers for flashing
			begin 
			ff_array [k] = 0;
			end			
		end

	assign scb_data = scb[scb_addr];		// get the scb_data from the array
////////////////////////////////////////////////////////////////////////////////
// put scan control bytes in registers & palettes into odd/even registers for
// for combining into 12 bit wide entry into BRAM
// also cb2flash data is put in registers
	always @(posedge clk50)
		begin
		if (video_write_SRAM && write_is_scb) scb [video_wr_addr[7:0]] <= write_data; 
		if(video_write_SRAM && write_is_pal && write_is_even) even_pal <= write_data;
		if(video_write_SRAM && write_is_pal && ~write_is_even) odd_pal <= write_data[3:0];
// high if data is in flashing area 
		if(write_vid_is_ff) ff_array[video_wr_addr[7:0]] <= write_data;  
		end
////////////////////////////////////////////////////////////////////////////////
// reset or DCM lock out
	wire reset = ~reset_s | ~LOCKED_OUT;	
////////////////////////////////////////////////////////////////////////////////
//	settings from the external flash			 
	reg [7:0] txt_bk = 8'hF0;	// holds text text color hi 4bits - bkgnd color lo 4bits
	reg [3:0] bdr = 4'h0;		// holds border color
	reg bw_dhr = 0;				// black & white dhr
	reg [7:0] CBVID = 8'h01;	// CB video is softswitch $C02A 
										// bit 0 = 1 for SHR640 dither
										// bit 1 = 1 for two grays in low res
	reg init_done = 0;			// 1 if settings routine is done 
////////////////////////////////////////////////////////////////////////////////
// instantiate the video module (vga_video.v)	
// labels ending in _s are synchronized with the main clock
   vga_video 
     video_module (						
		   .clk50(clk50), 					// 50 MHz main clock from DCM
		   .char_data(char_data), 			// from character ROM 
		   .ram_data(video_ram_data), 	// from video data SRAM or BRAM
		   .bus_addr(bus_a_s), 				// Apple slot bus address
			.bus_data(bus_d_s),				// Apple slot bus data
		   .ph1_s(ph1_s),						// Apple phase1 clock
			.Q3(Q3_s),							// Apple Q3 clock
		   .rw(BUS_RW), 						// read/write from Apple slot bus
		   .reset_n(~reset), 				// Apple reset
			.devsel_s(devsel_s),				// Apple slot /DEVSEL 	 
`ifdef bbug			
			.bugin(bugin),						// pass signal for bug border debugging
`endif
			.init_done(init_done),			// read saved settings routine done = 1
			.txt_bk(txt_bk),					// text & background color settings
			.bdr(bdr),							// border color setting
			.bw_dhr(bw_dhr),					// grayscale DHR setting
			.cbvideo(CBVID),						// dither 640 SHR & 2 grays settings
			.scb_data(scb_data),				// input scanline from BRAMwith
			.scb_addr(scb_addr),				// output scanline address in BRAM
			.pal_data(pal_data),				// input 12 bit palette from BRAM
			.pal_addr(pal_addr),				// output palette even byte address in BRAM
			
		   .update_time(update_time),		// when it's OK to get Apple data from bus
			.auxmem(auxmem),					// high when mirroring auxram (bank1) to CB RAM
		   .v_sync(v_sync), 					// vertical sync for hdmi
		   .h_sync(h_sync),					// horizontal sync for hdmi
			.RED(RED),							// 8 bits for 24 bit hdmi & VGA
			.GREEN(GREEN),						// 8 bits for 24 bit hdmi & VGA
			.BLUE(BLUE),						// 8 bits for 24 bit hdmi & VGA
			.FRAME_ENABLE(frame_en),		// high when pixels are drawn			
			.PIX_CLK(PIX_CLK),				// 25 MHz for 640x480			
		   .rom_addr(rom_addr), 			// address used to get character data
		   .ram_address(ram_address),		// address to get ram_data from CB RAM buffer
			.clk25(clk25),						// half of 50 MHz main clock
			.io_read(io_read), 				// output 1 when reading i/o location
			.io_reg(io_reg)   				// output contains data to put on io bus
		   );

////////////////////////////////////////////////////////////////////////////////
// instantiate distributed ROM for text characters 	
   charROM charROM (
	       .a(rom_addr),						// address from vga_video.v
	       .spo(char_data),					// character data for text in vga_video.v
	       .clk(clk50)						// 50 MHz clock for reading chars
	       );
////////////////////////////////////////////////////////////////////////////////
// build_stamp inputs to cb2.v
	wire[31:0] build_date;
	wire[31:0] build_time;
////////////////////////////////////////////////////////////////////////////////
// instantiate build_stamp.v

	build_stamp build_stamp (
					.build_date (build_date),
					.build_time (build_time)
					);
////////////////////////////////////////////////////////////////////////////////
//	spi
	wire sck;						// spi controller clock
	reg mosi = 0;					// from spi controller
	assign FPGA_MOSI = mosi;	// to spi_sdi (serial data into device)
   assign FPGA_SPICLK = sck;	// to spi_sck (clock for device)

	wire [7:0] spi_out;			
	wire [7:0] spi_in;
	wire write_spi;
	wire spi_done;
////////////////////////////////////////////////////////////////////////////////	
// The following spi code is taken from the original JAT spi.v module.
// See copyright and license below:
// ====================================================================
//                         VECTOR-06C FPGA REPLICA
//
// 		Copyright (C) 2007, Viacheslav Slavinsky
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without modification, 
// are permitted provided that the following conditions are met:
//
//	Redistributions of source code must retain the above copyright notice, this 
//	list of conditions and the following disclaimer.
//
//	Redistributions in binary form must reproduce the above copyright notice, 
//	this list of conditions and the following disclaimer in the documentation 
//	and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
////////////////////////////////////////////////////////////////////////////////
 
   wire	miso = FPGA_MISO;		// serial data from device
   wire [7:0] di = spi_in;		// 8 bit parallel data in
   wire	wr = write_spi;		// 1 = write to spi
   wire [7:0] do;					// 8 bit parallel data out
	assign spi_out = do;
   reg	done;						// 1 = 8 bits shifted in/out
	assign spi_done = done;
   reg [7:0] shiftreg;
   reg [7:0] shiftski;
   reg [1:0] state = 0;
   reg scken = 0;

   assign sck = clk25_b & scken;	// scken used to switch on/off device clock
   assign do  = shiftreg;

   always @(posedge clk50 or negedge reset_s) 
		begin
      if (!reset_s) 
			begin
			state <= 0;			// reset state
			mosi <= 1'b0;		
			done  <= 0;			// not done
			scken <= 0;			// device clock off
			end 
		else 
			begin
			if (clk25_b)		// every other time 
				begin
				case (state)
				0:	begin
					if (wr) 
						begin
						done <= 1'b0;			
						state <= 1;					
						shiftreg <= di;				// load the data
						shiftski <= 8'b11111111;	// use as counter
						end
					end
				1: begin
					scken <= 1;							// device clock on
					mosi <= shiftreg[7];				// send a bit to device
					shiftreg <= {shiftreg[6:0],miso};	// left shift data & shift bit in from device
					shiftski <= {1'b0,shiftski[7:1]};	// right shift counter
					if (|shiftski == 0)	
						begin 							// counter is 0
						state <= 2;					
						scken <= 0;						// device clock off
						end
					end
				2:	begin
		//			mosi <= 1'b0; // shouldn't be necessary but a nice debug view (VS)
					done <= 1'b1;
					state <= 0;
					end
				default: ;
				endcase
			end
      end
   end
// end spi
////////////////////////////////////////////////////////////////////////////////
// peripheral I/O registers
// used as communications between Apple & CB - see c3000_notes
	reg [7:0] C0n0 = 0;
	reg [7:0] C0n1 = 0;
	reg [7:0] C0n2 = 0;
	reg [7:0] C0n3 = 0;
	reg [7:0] C0n4 = 0;
	reg [7:0] C0n5 = 0;
	reg [7:0] C0n6 = 0;
	reg [7:0] C0n7 = 0;
	reg [7:0] C0nC = 0;
	reg [7:0] C0nF = 8'hCB;			// ID byte read only
	reg [7:0] bus_reg = 0;			// register for Apple read
////////////////////////////////////////////////////////////////////////////////
// Apple read/write signal sync'd with CB clock
	reg rw_s = 0;	
////////////////////////////////////////////////////////////////////////////////
// device IO (device_io)
// high when Apple /DEVSEL is low and phase 1 is high	
	wire device_io = (devsel_s == 1'b0) && (ph1_s == 1'b0);
// device read (dev_read)
// high when it's OK to read $C0nx area device I/O	
	wire dev_read = device_io & rw_s;
// device write (dev_write)
// high when it's OK to write $C0nx area device I/O 
	wire dev_write = device_io & ~rw_s;
////////////////////////////////////////////////////////////////////////////////
// SPI - M25 registers	
	reg write_spi_reg = 0;
	reg write_spi_reg4 = 0;			// flashing
	reg write_spi_reg6 = 0;			// read back error check
	reg [7:0] spi_in_reg = 0;
	reg [7:0] spi_in_reg4 = 0;
	reg [7:0] spi_in_reg6 = 0;
	wire one_byte_cmd;
	wire two_byte_cmd;
	reg write_cmd_reg = 0;
	reg [7:0] cmd_reg = 0;
	reg ff_on = 0;						// flashing on/off
	reg fr_on = 0;						// fast read on/off

	wire read_cmds;
////////////////////////////////////////////////////////////////////////////////
// SPI - M25 logic 

	assign write_spi = (one_byte_cmd || two_byte_cmd) ? write_cmd_reg : (ff_on) ? write_spi_reg4 : fr_on ? write_spi_reg6 : write_spi_reg;
	assign spi_in = (one_byte_cmd || two_byte_cmd) ? cmd_reg : ff_on ? spi_in_reg4 : fr_on ? spi_in_reg6 : spi_in_reg;

	assign read_cmds = (C0n1 == 8'h03) || (C0n1 == 8'h05);
	reg do_nvram = 0;

	reg cmd_cs = 1;
	reg spi_cs = 1;
	reg spi_cs4 = 1;					// flashing page program
	reg spi_cs6 = 1;					// read back error check
	reg [7:0] page_errs = 0;
	reg [4:0] nvram_state = 0;
	reg [4:0] nvram_state4 = 0;  	// flashing 
	reg [4:0] nvram_state6 = 0;  	// read back error check 
	
	always @(posedge clk50)
		begin
		clk25_b <= ~clk25_b;
		if (clk25_b) FLASH_CS <= (one_byte_cmd || two_byte_cmd) ? cmd_cs : ff_on ? spi_cs4 : fr_on ? spi_cs6 : spi_cs;
		end

////////////////////////////////////////////////////////////////////////////////
// read and write device I/O 	
	always @(negedge Q3_s)	
		begin
			if (do_nvram == 1'b1) do_nvram <= 1'b0;	// this keeps it to 1 Q3 clock cycle
			if((FLASH_CS == 1'b1)) 
				begin
				do_nvram <= 1'b0;
				if ((C0n0 == 8'h02) && read_cmds) C0n2 <= spi_out;
				if ((C0n0 == 8'h06))
					begin
						C0n7 <= page_errs;
						C0n2 <= nvram_state6;	
					end					
				end
// device I/O write
		if(dev_write)						// test for write to device 
		 	begin
			case(bus_a_s[3:0])
			4'h0:								// commands 0 = reset regs, 1 = version, 2 = M25,
												// 3 - do nothing, 4 - flashing, 6 - error test
				begin
				C0n0 <= bus_d_s;  		// make it stick
				if(bus_d_s == 8'h00)		// reset io regs command
					begin
					C0n1 <= 8'h00;			// yes, these are the only regs reset to zero
					C0n3 <= 8'h00;
					C0n4 <= 8'h00;			// note, C0n2 is not reset
					C0n5 <= 8'h00;
					C0n6 <= 8'h00;
					C0n7 <= 8'h00;
					C0nC <= 8'h00;
					end
				if(bus_d_s == 8'h01)		// version command
					begin
					C0n1 <= min_version;	// 0 - 255
					C0n2 <= maj_version; // 0 - 254 (255 is used for test versions)
					end					
				if(bus_d_s == 8'h09)		// build date command
					begin
					C0n3 <= build_date[7:0];	// 
					C0n4 <= build_date[15:8];	//
					C0n5 <= build_date[23:16];	// 
					C0n6 <= build_date[31:24];	//					
					end
				if(bus_d_s == 8'h0A)		// build date command
					begin
					C0n3 <= build_time[7:0];	// 
					C0n4 <= build_time[15:8];	//
					C0n5 <= build_time[23:16];	// 
					C0n6 <= build_time[31:24];	//					
					end
				end

			4'h1: C0n1 <= bus_d_s;		// usually data low byte
			4'h2: C0n2 <= bus_d_s;		// usually data high byte
			4'h3:	begin						// strobe for read/write of M25 or flashing 
												// flashing M25 or error checking M25
					C0n3 <= bus_d_s;

					if((C0n0 == 8'h02) || (C0n0 == 8'h04) || (C0n0 == 8'h06))
						begin						
						do_nvram <= 1'b1;	// starts execution of M25 code
						end
					end			
			4'h4: C0n4 <= bus_d_s;		// address low	
			4'h5: C0n5 <= bus_d_s;		// address mid
			4'h6: C0n6 <= bus_d_s;		// address high
			4'h7: C0n7 <= bus_d_s;		// page errors (flashing)
			4'hC: C0nC <= bus_d_s;			
			endcase
			end
// device I/O read
		if(dev_read)						// register data for reading
			begin								// from the Apple II
			case(bus_a_s[3:0])
			4'h0: bus_reg <= C0n0; 		// command
			4'h1: bus_reg <= C0n1;		// data		 
			4'h2: bus_reg <= C0n2; 		// data
			4'h3: begin
					bus_reg <= C0n3;		// m25 utilities
// flashing				
					if((C0n0 == 8'h02) || (C0n0 == 8'h04) || (C0n0 == 8'h06)) do_nvram <= 1'b1;
					end
			4'h4: bus_reg <= C0n4; 		// address low
			4'h5:	bus_reg <= C0n5; 		// address mid
			4'h6: bus_reg <= C0n6;		// address high
			4'h7: bus_reg <= C0n7;		// page errors
			4'hC: bus_reg <= C0nC;
			4'hF:	bus_reg <= C0nF;		// always CB
			endcase
			end
		end
////////////////////////////////////////////////////////////////////////////////
// put data on the bus for reading in the device IO area and softswitches
	assign BUS_D = dev_read ? bus_reg : io_read ? io_reg : 8'bz;

////////////////////////////////////////////////////////////////////////////////
//	initial set-up from flash RAM data (run once)
// uses read routine below
// a program like cb2config can read/write settings that are stored in flash RAM
// the format of the data is:
// bytes $3 to $6 are ID bytes to insure that we have the right data
// byte:
//			$10 text and background colors
//			$11 border color (bits 0-3)
//			$12 monochrome Double Hires Graphics 
//			$13 CBvideo options: bit 0 = 1 for dither 640 SHR, 
//										bit 1 = 1 for two grays in lowres graphics
//			$16 reserved
  
reg [3:0] init_state = 0; 
reg init_set = 1;				// routine runs only once at startup
reg [7:0] init_cnt = 8'h0; // so we can get up to 256 bytes

always @(posedge clk25_b)
	begin
	if(init_set & LOCKED_OUT)
		begin
			if((init_state == 13) && (spi_done))
				begin
				case (init_cnt)
				8'h03: if (spi_out != 8'hC3) init_set <= 1'b0;	// $3 - $6 test for magic #
				8'h04: if (spi_out != 8'hC2) init_set <= 1'b0;
				8'h05: if (spi_out != 8'h53) init_set <= 1'b0;
				8'h06: if (spi_out != 8'h30) init_set <= 1'b0;
				8'h10: txt_bk <= spi_out;								// text & background colors
				8'h11: bdr <= spi_out[3:0];							// border colors
				8'h12: bw_dhr <= spi_out[5];							// monochrome DHR
				8'h13: CBVID <= spi_out;								// CB video settings
				8'hFF: begin
						 init_set <= 1'b0;
						 init_done <= 1'b1;
						 end
				default:;
				endcase
				end
		end
	end
////////////////////////////////////////////////////////////////////////////////
// Main commands called from the Apple II
//	read, program or sector erase NVRAM (M25)
// Read Data Bytes (READ)	$03
//	Page Program	 (PP)		$02
//	Sector Erase	 (SE)		$D8

always @(posedge clk25_b or negedge reset_s)
	begin
	if(reset_s == 1'b0)						// reset
		begin
		spi_cs <= 1'b1;
		write_spi_reg <= 1'b0;
		nvram_state <= 0;
		end
	else if (init_set)						// initial set-up runs once
		begin
		case (init_state)
		0: begin
			init_state <= 1;
			spi_cs <= 1'b0;					// turn on M25
			spi_in_reg <= 8'h03;				// read
			write_spi_reg <= 1'b1;			// write command seq		
			end
		1: init_state <= 2;	
		2: begin										
			write_spi_reg <= 1'b0;				
			if (spi_done) 
				begin
				init_state <= 3;
				end
			end
		3: begin										
			spi_in_reg <= 8'h08;					// write high addr byte
			init_state <= 4;
			write_spi_reg <= 1'b1;
			end
		4: init_state <= 5;
		5: begin										
			write_spi_reg <= 1'b0;
			if (spi_done) 
				begin
				init_state <= 6;
				end
			end
		6: begin									
			spi_in_reg <= 8'h00;					// write mid addr byte
			init_state <= 7;
			write_spi_reg <= 1'b1;
			end
		7: init_state <= 8;
		8: begin										
			write_spi_reg <= 1'b0;
			if (spi_done) 
				begin
				init_state <= 9;
				end
			end
		9: begin							
			spi_in_reg <= init_cnt;				// write low addr byte
			init_state <= 10;		
			write_spi_reg <= 1'b1;
			end
		10: init_state <= 11;
		11:begin									
			write_spi_reg <= 1'b0;			// stop writing
			if (spi_done)
				begin
				spi_in_reg <= 8'h00;
				init_state <= 12;
				write_spi_reg <= 1'b1;
				end
			end
		12: init_state <= 13;
		13:begin
			write_spi_reg <= 1'b0;		
			if (spi_done)
				begin
				spi_cs <= 1'b1;				// turn off M25
				init_state <= 14;
				end
			end
		14:begin
			init_cnt <= init_cnt + 1;
			init_state <= 0;
			end
		default: ;
		endcase
		end										// end of initial settings
// read - sector erase - slow flash - one byte at a time	
	else if(((C0n1 == 8'h03) || (C0n1 == 8'h02) || (C0n1 == 8'hD8)) && (C0n0 == 8'h02))    
		begin
		case (nvram_state)
		0: begin										
			if(do_nvram) 
				begin
				nvram_state <= 1;
				spi_cs <= 1'b0;					// turn on M25
				spi_in_reg <= C0n1;				// read, program or sector erase
				write_spi_reg <= 1'b1;			// write command seq
				end
			end
		1: nvram_state <= 2;	
		2: begin										
			write_spi_reg <= 1'b0;				
			if (spi_done) 
				begin
				nvram_state <= 3;
				end
			end
		3: begin										
			spi_in_reg <= C0n6;					// write high addr byte
			nvram_state <= 4;
			write_spi_reg <= 1'b1;
			end
		4: nvram_state <= 5;
		5: begin										
			write_spi_reg <= 1'b0;
			if (spi_done) 
				begin
				nvram_state <= 6;
				end
			end
		6: begin									
			spi_in_reg <= C0n5;					// write mid addr byte
			nvram_state <= 7;
			write_spi_reg <= 1'b1;
			end
		7: nvram_state <= 8;
		8: begin										
			write_spi_reg <= 1'b0;
			if (spi_done) 
				begin
				nvram_state <= 9;
				end
			end
		9: begin							
			spi_in_reg <= C0n4;				// write low addr byte
			nvram_state <= 10;		
			write_spi_reg <= 1'b1;
			end
		10: nvram_state <= 11;
		11:begin									
			write_spi_reg <= 1'b0;			// stop writing
			if (spi_done)
				begin
				if (C0n1 == 8'hD8)			// if sector erase
					begin
					spi_cs <= 1'b1;			// turn off M25
					nvram_state <= 0;					
					end
				else
					begin
					spi_in_reg <= (C0n1 == 8'h02) ? C0n2 : 8'h00;
					nvram_state <= 12;
					write_spi_reg <= 1'b1;
					end
				end
			end
		12: nvram_state <= 13;
		13:begin
			write_spi_reg <= 1'b0;		
			if (spi_done)
				begin
				spi_cs <= 1'b1;				// turn off M25
				nvram_state <= 0;
				end
			end
		default: ;
		endcase
		end
	end
////////////////////////////////////////////////////////////////////////////////
// flash from the Apple II
// flashing (was fastflash)	
// uses page program to flash 256 bytes from SRAM

	reg[8:0] ff_addr = 9'h000;		// flashing uses hgr2 addr 
	reg[8:0] fr_addr = 9'h000;		// read back & error check

always @(posedge clk25_b or negedge reset_s)
	begin
	if(reset_s == 1'b0)						// reset
		begin
		spi_cs4 <= 1'b1;
		write_spi_reg4 <= 1'b0;
		nvram_state4 <= 0;
		ff_addr <= 9'h000;
		ff_on  <= 1'b0;
		end
	else if  ((C0n1 == 8'h02) && (C0n0 == 8'h04))		// flash using page program
		begin
		case (nvram_state4)
		0: begin										
			if(do_nvram) 
				begin
				ff_on <= 1'b1;
				nvram_state4 <= 1;
				spi_cs4 <= 1'b0;					// turn on M25
				spi_in_reg4 <= C0n1;				// program only 
				write_spi_reg4 <= 1'b1;			// write command seq
				end
			end
		1: nvram_state4 <= 2;	
		2: begin										
			write_spi_reg4 <= 1'b0;				
			if (spi_done) 
				begin
				nvram_state4 <= 3;
				end
			end
		3: begin										
			spi_in_reg4 <= C0n6;					// write high addr byte
			nvram_state4 <= 4;
			write_spi_reg4 <= 1'b1;
			end
		4: nvram_state4 <= 5;
		5: begin										
			write_spi_reg4 <= 1'b0;
			if (spi_done) 
				begin
				nvram_state4 <= 6;
				end
			end
		6: begin									
			spi_in_reg4 <= C0n5;					// write mid addr byte
			nvram_state4 <= 7;
			write_spi_reg4 <= 1'b1;
			end
		7: nvram_state4 <= 8;
		8: begin										
			write_spi_reg4 <= 1'b0;
			if (spi_done) 
				begin
				nvram_state4 <= 9;
				end
			end
		9: begin							
			spi_in_reg4 <= 8'h00;				// write low addr byte (always zero)
			nvram_state4 <= 10;		
			write_spi_reg4 <= 1'b1;
			ff_addr <= 9'h000;
			end
		10: begin
			 nvram_state4 <= 11;
			 end
		11:begin	
			write_spi_reg4 <= 1'b0;			// stop writing
			if (spi_done)
				begin								// this is where we flash 256 bytes 
				spi_in_reg4 <= ff_array[ff_addr[7:0]];
				ff_addr <= ff_addr + 1;
				if (ff_addr == 9'h100) 
					begin
					nvram_state4 <= 12;
					ff_addr <= 9'h000;  		// reset		
					end
				else nvram_state4 <= 10;
				write_spi_reg4 <= 1'b1;
				end
			end
		12:begin
			nvram_state4 <= 13;			
			end
		13:begin
			write_spi_reg4 <= 1'b0;		
			if (spi_done)						
				begin			
				spi_cs4 <= 1'b1;				// turn off M25
				nvram_state4 <= 0;
				ff_on  <= 1'b0;				// flashing is off
				end
			end
		default: ;
		endcase
		end
	end
////////////////////////////////////////////////////////////////////////////////
//	 read back & test for errors 
always @(posedge clk25_b or negedge reset_s)
	begin
	if(reset_s == 1'b0)						// reset
		begin
		spi_cs6 <= 1'b1;
		write_spi_reg6 <= 1'b0;
		nvram_state6 <= 0;
		fr_addr <= 9'h000;
		fr_on  <= 1'b0;
		page_errs <= 8'h00;
		end
	else if  ((C0n1 == 8'h03) && (C0n0 == 8'h06))
		begin
		case (nvram_state6)
		0:	begin
			if(do_nvram) 
				begin
				fr_on <= 1'b1;
				fr_addr <= 9'h000;
				nvram_state6 <= 1;
				spi_in_reg6 <= 8'h05;		// read status 
				write_spi_reg6 <= 1'b1;
				spi_cs6 <= 1'b0;
				end
			end
		1: nvram_state6 <= 2;
		2: begin										
			write_spi_reg6 <= 1'b0;				
			if (spi_done) nvram_state6 <= 3;
			end
		3: begin
			spi_in_reg6 <= 8'h00;			// dummy byte
			nvram_state6 <= 4;
			write_spi_reg6 <= 1'b1;			// write command seq
			end
		4: nvram_state6 <= 5;
		5: begin										
			write_spi_reg6 <= 1'b0;				
			if (spi_done) 
				begin
				if(spi_out[0] == 1'b0) 
					begin
					spi_cs6 <= 1'b1;					
					nvram_state6 <= 6;
					page_errs <= 8'h00;
					end
				else 
					begin
					nvram_state <= 1;
					spi_in_reg6 <= 8'h05;	// read status 
					write_spi_reg6 <= 1'b1;					
					end
				end
			end
		6: begin
			spi_in_reg6 <= C0n1;				// read
			nvram_state6 <= 7;
			spi_cs6 <= 1'b0;
			write_spi_reg6 <= 1'b1;			// write command seq
			end
		7: nvram_state6 <= 8;
		8: begin										
			write_spi_reg6 <= 1'b0;				
			if (spi_done) nvram_state6 <= 9;
			end
		 9:begin										
			spi_in_reg6 <= C0n6;				// write high addr byte
			nvram_state6 <= 10;
			write_spi_reg6 <= 1'b1;
			end
		10: nvram_state6 <= 11;
		11:begin										
			write_spi_reg6 <= 1'b0;
			if (spi_done) nvram_state6 <= 12;
			end
		12:begin									
			spi_in_reg6 <= C0n5;				// write mid addr byte
			nvram_state6 <= 13;
			write_spi_reg6 <= 1'b1;
			end
		13:nvram_state6 <= 14;
		14:begin										
			write_spi_reg6 <= 1'b0;
			if (spi_done) nvram_state6 <= 15;
			end
		15:begin							
			spi_in_reg6 <= fr_addr[7:0];	// write low addr byte
			nvram_state6 <= 16;			
			write_spi_reg6 <= 1'b1;
			end
		16:nvram_state6 <= 17;
		17:begin
			write_spi_reg6 <= 1'b0;		
			if (spi_done)
				begin
				nvram_state6 <= 18;			
				write_spi_reg6 <= 1'b1;				
				end
			 end
		18:nvram_state6 <= 19;
		19:begin	
			write_spi_reg6 <= 1'b0;		
			if (spi_done)
				begin							// compare each byte of the 256
				if(spi_out != ff_array[fr_addr[7:0]]) page_errs <= page_errs + 1;
				fr_addr <= fr_addr + 1;
				if(fr_addr == 9'h100) 
					begin
					fr_addr <= 9'h000;
					nvram_state6 <= 20;
					end
				else 
					begin
					nvram_state6 <= 6; 	
					spi_cs6 <= 1'b1;
					end
				end
			end
		20:begin
				begin
				spi_cs6 <= 1'b1;				// turn off M25
				nvram_state6 <= 0;
				fr_on <= 1'b0;
				end
			end
		default: ;
		endcase
		end
	end

////////////////////////////////////////////////////////////////////////////////
// M25 commands that can be called from the Apple
//
// one byte NVRAM commands
//
// Write enable (WREN) 	$06
//	Write disable (WRDI)	$04
//	Bulk Erase (BE)		$C7
//	Deep Power Down (DP)	$B9
//	Release Deep Power Down (RES)	$AB
//
// two byte NVRAM commands
//
// Read Status Register (RDSR)	$05
//	Write Status Register (WRSR)	$01

reg [2:0] cmd_state = 0;

assign one_byte_cmd = (C0n1 == 8'h06) || (C0n1 == 8'h04) || (C0n1 == 8'hC7) || (C0n1 == 8'hB9) || (C0n1 == 8'hAB) && (C0n0 == 8'h02);
assign two_byte_cmd = (C0n1 == 8'h05) || (C0n1 == 8'h01) && (C0n0 == 8'h02);
always @(posedge clk25_b or negedge reset_s)
	begin
	if(reset_s == 1'b0)
		begin
		cmd_cs <= 1'b1;
		cmd_state <= 0;
		write_cmd_reg <= 1'b0;
		end
	else if((C0n0 == 8'h02) && (one_byte_cmd || two_byte_cmd))
		begin
		case (cmd_state)
		0: begin									// use 1 cycle
			if (do_nvram)
				begin
				cmd_state <= 1;
				cmd_cs <= 1'b0;				// turn on M25
				cmd_reg <= C0n1;				// a one or two byte command
				write_cmd_reg <= 1'b1;
				end
			end
		1: cmd_state <= 2;
		2: begin
			write_cmd_reg <= 1'b0;			// stop writing
			if (spi_done)
				begin
				if (one_byte_cmd)
					begin
					cmd_cs <= 1'b1;			// turn off M25 if one byte command
					cmd_state <= 0;
					end
				else								// continue if reading status reg
					begin
					cmd_reg <= 8'h02;			// dummy
					cmd_state <= 3;
					write_cmd_reg <= 1'b1;
					end
				end
			end
		3: cmd_state <= 4;
		4: begin
			write_cmd_reg <= 1'b0;			// stop writing
			if (spi_done)
				begin		
				cmd_cs <= 1'b1;				// turn off M25 if reading status
				cmd_state <= 0;				
				end
			end
		default: ;
		endcase
		end
	end
////////////////////////////////////////////////////////////////////////////////
// Synchronize Apple slot signals to CB clock	
	reg rw_1 = 0;
	reg reset_1 = 0;
	reg phase0_1 = 0;
	reg Q3_1;
	reg devsel_1 = 0;
	reg iosel_1 = 0;
	reg iostrobe_1 = 0;
	reg ph1_1 = 0;
	
	always @(posedge clk50) 	
		begin
		rw_1 <= BUS_RW;
		rw_s <= rw_1;
      Q3_1 <= BUS_Q3; 
      Q3_s <= Q3_1;
		phase0_1 <= BUS_PHASE0;
		phase0_s <= phase0_1;
      ph1_1 <= BUS_PHASE1;
      ph1_s <= ph1_1;
		reset_1 <= BUS_RESET;
		reset_s <= reset_1;
		devsel_1 <= BUS_DEV_SEL;
		devsel_s <= devsel_1;
		bus_a_1 <= BUS_A;
		bus_a_s <= bus_a_1;
		bus_d_1 <= BUS_D;
		bus_d_s <= bus_d_1;
		end
////////////////////////////////////////////////////////////////////////////////
// bbug - used for debugging -- increments border color 
`ifdef bbug
	assign bugin = 1'b0;		// replace 1'b0 with a wire/reg to send state to bugborder
`endif 
endmodule
