////////////////////////////////////////////////////////////////////////////////
// c3000 project
//////////////////////////////////////////////////////////////////////////////////
// vga_video.v
// This is a module for the Carte Blanche II card under the top module cb2.v.
// Based on the video routines of Alex Freed's JAT (Jack of All Trades) project
// for the Carte Blanche I. 
// The following is the information for vga_video.v module for the CB II.  
//////////////////////////////////////////////////////////////////////////////////
// Its main functions are:
//
// Generate video output as VGA and/or HDMI.
// 	It does this by maintaining counters for horizontal and vertical positions 
// 	of the visible pixels and the blanking.  The horizontal and vertical sync
// 	pulses are also determined by those counters.
//
// 	It creates the pixels for display by processing the data from buffers in 
// 	the CB SRAM and/or FPGA BRAM. The video differs depending on the video 
//    mode. The resulting pixel data is sent to cb2.v for output.
//
// Determines the correct video mode (text or various graphics modes)
//	by watching the Apple slot bus for reads/writes to the Apple II softswitch
//	area at $C0xx.
// 
// Determine the correct timing for reading from or writing to the video buffers.
//  
//////////////////////////////////////////////////////////////////////////////////
// This versions output will work with either the HDMI or VGA.
// It is for an Apple //e only.  
// 
// The resolution is 640x480 @ 60 Hz. The DCM (main clock) output is 50 MHz and is divided
// down to 25 MHz for the VGA / HDMI pixel clock.
//////////////////////////////////////////////////////////////////////////////////
`include "c3000_defs.v"

module vga_video (
		  input clk50, 
		  input [7:0] char_data, 		// from character ROM
		  input [7:0] ram_data,  		// from video data SRAM thru cb2.v
		  input [15:0] bus_addr,		// Apple slot bus address - sync'd in cb2.v
		  input [7:0] bus_data,  		// Apple slot bus data - sync'd in cb1.v
		  input ph1_s,						// Apple phase1 clock - sync'd in cb2.v
		  input Q3,							// Apple Q3 clock - sync'd in cb2.v
		  input rw,							// read/write from Apple 
		  input reset_n,					// reset from Apple
		  input devsel_s,					// Apple slot /DEVSEL		
`ifdef bbug
			input bugin,					// use this to pass a signal for bugborder
`endif	
			input init_done,				// if settings recieved from M25 (external flash chip)
			input [7:0] txt_bk,			// text & background color settings from M25
			input [3:0] bdr,				// border color setting from M25 
			input bw_dhr,					// black & white dhr setting from M25
			input [7:0] cbvideo,			// CB video is softswitch $C02A 
												// bit 0 = 1 for SHR640 dither
												// bit 1 = 1 for two grays in low res
			input [7:0] scb_data,		// input scanline from BRAM
			output [7:0] scb_addr,		// output scanline address in BRAM
			input [11:0] pal_data,		// input palette color from BRAM
			output [7:0] pal_addr,		// output palette color address in BRAM
			output update_time,			// when it's OK to write Apple data to SRAM
			output reg auxmem,			// high when mirroring auxram (bank1) to CB RAM 
			output v_sync,					// for vga
			output h_sync,					// for vga
			output [7:0] RED ,			// 8 bits for 24 bit hdmi & VGA
			output [7:0] GREEN,			// 8 bits for 24 bit hdmi & VGA
			output [7:0] BLUE,			// 8 bits for 24 bit hdmi & VGA
			output FRAME_ENABLE,			// high when pixels are drawn			
			output PIX_CLK,				// 25 MHz for 640x480
			output reg [10:0] rom_addr,	// address used to get character data  
			output [16:0] ram_address, 	// address to get ram_data from CB RAM buffer
		  output clk25,					// half of 50 MHz clock
		  output reg io_read,			// high when Apple reads IO area
		  output reg [7:0] io_reg		// data when Apple reads IO area
		  );
///////////////////////////////////////////////////////////////////////////
// registers used for synchronizing Apple signals with the main CB clock //
	reg rw_s;
	reg rw_1;
	reg reset_s;
	reg reset_1;
///////////////////////////////////////////////////////////////////////////
// the two main counters for positions of pixels, blanking and sync signals
	reg [9:0]    h_count = 0;		// horz dot count 
   reg [9:0] 	 v_count = 47;		// vert line count
///////////////////////////////////////////////////////////////////////////
// 25 MHz clock derived from 50 MHz main clock
   reg clk1 = 0;						// used for pixel clock
///////////////////////////////////////////////////////////////////////////
// settings from softswitch area $C0xx
   reg TEXT = 1;						// TEXT mode initially on
   reg MIXED = 1;						// MIXED mode initially on (4 lines of text)
   reg PAGE2 = 0;						// PAGE 2 initially off
	reg HIRES = 0;						// High Res graphics initially off
   reg mode_80COL = 0;     		// 80 column mode initially off
	reg mode_80STORE = 0;			// 80 column store initially off
	reg AltCharSet = 0;				// alternate character set initially off
	reg WRITE_AUX = 0;      		// 1 = aux-same as $C005 on, 0 = main - same as $C004 on
	reg LINEAR_SHR = 0;				// $C029 bit 6
 	reg SHR = 0;						// Super Hires graphics initially off
	reg DHR = 0;						// Double Hires graphics initially off
	reg monochrome = 0;				// monochrome (grayscale) Double Hires graphics initially off
	reg [3:0] BORDER_COLOR_NUM = 4'b0000;  // border init to black
	reg [3:0] TEXT_COLOR_NUM = 4'b1111;		// text init to white
	reg [3:0] BACK_COLOR_NUM = 4'b0000;		// background init to black
	reg [7:0] CBVID = 8'h01;		// CB video is softswitch $C02A 
											// bit 0 = 1 for SHR640 dither
											// bit 1 = 1 for two grays in low res
///////////////////////////////////////////////////////////////////////////
// flags for border, text & background from softswitches											
	reg border_flag = 0;				// 1 if use settings from softswitch 
	reg tb_flag = 0;					// 1 if use text & background settings from softswitch

///////////////////////////////////////////////////////////////////////////
// bugborder uses VGA border color as data tester
`ifdef bbug
	reg bugborder = 0;		// added 9/30/10 for testing wires & single bit regs
`endif
///////////////////////////////////////////////////////////////////////////
// tester allows a softswitch register ($C042) to be used to pass a byte
// for testing different values in a working bitstream 
`ifdef tester
	reg[7:0] test42 = 0;		// input for various tests
`endif
///////////////////////////////////////////////////////////////////////////
	reg [7:0] 	 shifter7;						// shifter used in non-super hires modes
   wire [2:0] 	 scanline = v_count[3:1];	// used for determining video address
   reg [2:0] 	 count7;							// 0-6 counter used with shifter7
   reg [6:0] 	 hc = 0;							// horizontal character count
   reg [4:0] 	 vc = 0;							// vertical character count
	reg [4:0] 	count28 = 0; 					// 0-27 counter added for DHR
	reg [7:0]	shr_bytes = 0; 	 			// super hires bytes in 1 scanline	
   reg [15:0] shr_addr = 16'h2000;			// address of bytes for super hires	- 1
	wire disp40_active;							// on when visible part of display is active
	wire border_active;							// on when border is displayed
	wire hor_active;								// on when horizontal part of display is active
	wire vert_active;								// on when vertical part of display is active
   reg hor_active_reg;							// gets set to hor_active (line being drawn) 
	wire shift_bitsum = disp40_active ? shifter7 > 0 : 1'b0;	// used with Hires 
   wire GR =  ~(TEXT | ( MIXED & vc[4] & vc[2])); // not text or mixed and >= line 20 
   reg  GR_r, GR_r2;								// used to delay graphics
	wire hgr;										// hires graphics (wit logic assigned)
///////////////////////////////////////////////////////////////////////////
// colors_12 submodules (16 colors)
// registers are inputs - wires are outputs
// HIRES and Super HIRES do not use these submodules
// Low Resolution	
   reg [3:0] 	low_color_nibble;		// used to select lores color (16 colors)
   wire [11:0] lowres_color_val12;	// actual 12 bit color
// border, text & background colors
	reg [3:0] border_number;			// used to select border color (16 colors)
	wire [11:0] border_color12;		// actual 12 bit color (border)
	reg [3:0] text_number;				// used to select text color (16 colors)
	wire [11:0] text_color12;			// actual 12 bit color (text)
	reg [3:0] back_number;				// used to select background color (16 colors)
	wire [11:0] back_color12;			// actual 12 bit color (background)
// hires colors
	reg [11:0] color12;					// hires 12 bit color
// double hires colors
	wire [11:0]	dhr_color12;			// double hires color 12 bit
	reg [3:0] dhrcolor_number;			// used to select double hires color (16 colors) 
// Double Low Resolution colors
	wire[3:0] dlr_nibble; 			  	// dlr uses dhr palette
	wire [11:0] dlr_color12;			// double lores 12 bit color	
///////////////////////////////////////////////////////////////////////////	
	wire	dhires;							// double hires
   reg 	hor_del2 = 0;					// delays horizontal active signal
   reg [4:0] flash_count = 0;   		// counter for FLASH text speed
   reg flash = 1;							// FLASH text mode
	reg inverse = 0;						// INVERSE text mode
   wire flashing = flash ? flash_count[4] : 1'b0; // FLASH on and off      
	wire dith640;							// 1 if SHR 640 mode is dithered	
	wire Q3_s = Q3;						// wire Q3 synchronized with main clock from cb2.v									
   reg 	vid_del;							// video delay
	assign clk25 = clk1;					// send clk1 to cb2.v
	assign PIX_CLK = clk1;				// pixel clock
	assign dhires = DHR & mode_80COL & GR & ~( MIXED & vc[4] & vc[2]);	// double hires, 80 column and graphics on	
///////////////////////////////////////////////////////////////////////////
// create horizontal sync & vertical sync for video output
	assign 	    h_sync = (h_count < 659) || (h_count > 755);
	assign 	    v_sync = (v_count < 493) || (v_count > 494);
///////////////////////////////////////////////////////////////////////////
// active part of screen (not borders)

	assign hor_active = (SHR == 1) ? ((h_count >= dith640) & (h_count < 641)) : ~mode_80COL ? (h_count > 24) & (h_count < 592) :  (h_count > 33) & (h_count < 601);		
	assign vert_active =  (SHR == 1) ? (v_count > 47) & ( v_count < 448) : (v_count > 47) & ( v_count < 432); 
//////////////////////////////////////////////////////////////////////////////////	
// display delay
// used to delay the active drawing area of the screen when in SHR
// this helps position the display the same in both HDMI & VGA  
	reg[4:0] display_delay = 0;

	always @(posedge clk1) 
		begin
		display_delay[4:0] <= {display_delay[3:0], (hor_active_reg & vert_active)}; 
		end
///////////////////////////////////////////////////////////////////////////
// active part of screen including delays (not borders)		
	assign disp40_active = SHR ? display_delay[0] : hor_active_reg & hor_del2 & vert_active;
//////////////////////////////////////////////////////////////////////////////////
// border active
	assign border_active = (h_count < 640) & (v_count < 480) & ~disp40_active;
//////////////////////////////////////////////////////////////////////////////////	
// determine when it's OK to read and write RAM buffers (in cb1.v)
// update_time is for writing to the buffers
	assign update_time = (hgr) ?  ((count7 == 0) & (h_count[0] == 1'b1)) : (h_count[1] == 1'b0);  
//////////////////////////////////////////////////////////////////////////////////		
// HDMI & VGA pixels
	reg aux_byte = 0;						// 1 if data from aux memory
	reg[11:0] hdmi2 = 0;					// 12 bit color used for both 24 bit hdmi and 24 bit VGA
	reg[11:0] shr_color12 = 0;			// 12 bit Super Hires color
//////////////////////////////////////////////////////////////////////////////////		
// text / background pixels on/off	
	wire vp = (~FRAME_ENABLE) ? 0 : (shifter7[0] ^ (inverse ^ flashing)); 
//////////////////////////////////////////////////////////////////////////////////		
// TEXT color (12 bit) with color background
	wire [11:0] text12 = (hor_active_reg & hor_del2) ? (({vp,vp,vp,vp,vp,vp,vp,vp,vp,vp,vp,vp} & text_color12[11:0]) ^ ({~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp} & back_color12[11:0])) : 12'b000;
//////////////////////////////////////////////////////////////////////////////////		
//	Graphics color (12 bit) not including SHR
	wire [11:0] GR_RGB12 = (dhires && HIRES) ? dhr_color12 : (dhires && ~HIRES & ~aux_byte) ? dlr_color12 : (HIRES == 0) ? lowres_color_val12 : color12; 


//////////////////////////////////////////////////////////////////////////////////		
// frame enable needed for HDMI - VGA chips
	assign FRAME_ENABLE = (h_count > 2) & (h_count < 641) & (v_count < 480);
//////////////////////////////////////////////////////////////////////////////////		
// when drawing to screen get the colors depending on the video mode	
	always @(posedge clk1)
		begin
		hdmi2 <= (border_active) ? border_color12 : (SHR & disp40_active) ? shr_color12 : (GR_r2 & disp40_active) ? GR_RGB12 : text12 ;
		end											
//////////////////////////////////////////////////////////////////////////////////				
// convert the 12 bit colors to 24 bits for hdmi output by repeating each color's 4 bit sequence
	assign RED = {hdmi2[11],hdmi2[10],hdmi2[9],hdmi2[8],hdmi2[11],hdmi2[10],hdmi2[9],hdmi2[8] };
	assign GREEN = {hdmi2[7],hdmi2[6],hdmi2[5],hdmi2[4],hdmi2[7],hdmi2[6],hdmi2[5],hdmi2[4]};
	assign BLUE = {hdmi2[3],hdmi2[2],hdmi2[1],hdmi2[0],hdmi2[3],hdmi2[2],hdmi2[1],hdmi2[0]};	
//////////////////////////////////////////////////////////////////////////////////
// colors_12.v submodule instanstiation (12 bit color tables)
	colors_12 colors_12 (
		.monochrome(monochrome),						// input
		.CBVID(CBVID),										// input
		.border_number(border_number),				// input
		.text_number(text_number),						// input
		.back_number(back_number),						// input
		.dhrcolor_number(dhrcolor_number),			// input	
		.nibble(low_color_nibble),						// input
		.dlr_nibble(dlr_nibble),						// input
		.border_color12(border_color12),				// output
		.text_color12(text_color12),					// output
		.back_color12(back_color12),					// output
		.dhr_color12(dhr_color12),						// output
		.lowres_color_val12(lowres_color_val12),	// output
		.dlr_color12(dlr_color12)						// output
			);
//////////////////////////////////////////////////////////////////////////////////	
// okay to shift video bits
	wire allow_shift7 = (mode_80STORE & mode_80COL | DHR) ? 1'b1: h_count[0]; 
//////////////////////////////////////////////////////////////////////////////////	
// creates the pixel clock.
	always @(posedge clk50) clk1 <= ~clk1;			// divides clk50 by 2 - pixelclock
//////////////////////////////////////////////////////////////////////////////////	
// increments through the dots
//////////////////////////////////////////////////////////////////////////////////	
// increments the counters used for positioning the pixel output and sync pulses.
// increments counters used in calculating the pixel data and shifting.
// calculates the ROM address for TEXT bits.
	reg line_dbl = 0;
	
   always @(posedge clk50) 
		begin
		if(clk1) 													
			begin				
// SHR address calc & line doubler
			if(hor_active & vert_active)			// only if display is active
				begin										
				if (h_count[1:0] == 2'b01)			// only on every 4th count starting with 1
															// on IIgs code this is starting with 0
					begin
					shr_addr <= shr_addr + 1;		// increment the address counter
					if(shr_addr == 16'h9CFF) shr_addr <= 16'h2000;	// restart counter
					shr_bytes <= shr_bytes + 1;	// increment byte in line counter							
					if (shr_bytes == 8'h9F) 
						begin
						if (line_dbl == 0) shr_addr <= shr_addr - shr_bytes; // line doubler
						line_dbl <= ~line_dbl; 		// flip doubler flag
						end
					end
				end	// if(hor_active & vert_active)
// DHR counter
`ifdef tester
			if ((h_count < 8 + test42)  || (count28 == 27)) count28 <= 0;
`else
			if ((h_count < 8)  || (count28 == 27)) count28 <= 0;
`endif
			else count28 <= count28 + 1;
// reset counters
			if (h_count == 799) 							// end of horizontal count 0-799  
				begin
				h_count <= 0; 								// start over
				count7 <= 0;  								// reset 0-6 counter
				shr_bytes <= 8'h00;
// SHR vertical position
				if ((v_count == 447) || (v_count < 48)) 
					begin
					shr_addr <= 16'h2000;
					line_dbl <= 1'b0;
					end
// line counter
				if(v_count == 524) v_count <= 0;		// ( last line 0-524 start over)
				else v_count <= v_count +1;			// not last line so start next line
				end  						
// horizontal counter
			else 												// not end of horizontal count 0-799
				begin
				h_count <= h_count + 1;					// increment the horizontal count 0-799
// counter for shifting
				if(allow_shift7) 							// every clk1 for 80c, every other clk1 for 40c
					begin
					if(count7 == 6) count7 <= 0;		// end of 0 to 6 count so reset 
					else count7 <= count7 + 1;			// not end so increment	
					end    
				end
			end 				// if clk1
// character ROM address
		else 					// not clk1 send character address to cb2.v 		
			if(count7 == 3'b0) rom_addr <= {ram_data[7], (ram_data[6] & AltCharSet) | (ram_data[7] & ram_data[6]), ram_data[5:0], scanline};		
		end // posedge clk50
//////////////////////////////////////////////////////////////////////////////////			
// Aux memory - determine when to read or write aux memory (bank1)
	reg load_aux_mem = 0;
	always @(posedge clk50)
		begin
		load_aux_mem <= (mode_80COL && (TEXT | ( MIXED & vc[4] & vc[2]) | DHR) && aux_byte) | SHR;
		auxmem <= (mode_80STORE & PAGE2) | WRITE_AUX;  
		end
//////////////////////////////////////////////////////////////////////////////////		
// calculate video RAM addresses	
	wire X_PAGE2 = PAGE2 & ~mode_80STORE; 	
   reg 	high_bit;
	wire [5:0] high_vid_addr = (hgr) ? {1'b0, X_PAGE2, ~X_PAGE2, scanline} : {4'b0000, X_PAGE2, ~X_PAGE2};	// line
	assign hgr = (HIRES & GR & ~( MIXED & vc[4] & vc[2]));						  
	wire [3:0] addr_sum = {~hc[5],~hc[5],hc[4],hc[3]} +  {vc[4],vc[3],vc[4],vc[3]} + 1;		
	wire [15:0] apple_video_addr = (SHR) ? shr_addr : {high_vid_addr, vc[2:0], addr_sum ,hc[2:0]};
   reg [15:0]  apple_video_addr_reg;
	
	assign ram_address = {load_aux_mem, apple_video_addr_reg};	// this is the one used for getting ram_data from buffer 
//////////////////////////////////////////////////////////////////////////////////	
// delays and 
// border, text & background colors
	reg hor_del1 = 0;
	
	always @(posedge clk50)
		begin

		if (~disp40_active) 
			begin
`ifdef bbug				
			border_number <= border_flag ? BORDER_COLOR_NUM + bugborder : init_done ? bdr + bugborder : BORDER_COLOR_NUM;
`else				
			border_number <= border_flag ? BORDER_COLOR_NUM : init_done ? bdr : BORDER_COLOR_NUM;
`endif
			text_number   <= tb_flag ? TEXT_COLOR_NUM : init_done ? txt_bk[7:4] : TEXT_COLOR_NUM;		
			back_number	  <= tb_flag ? BACK_COLOR_NUM : init_done ? txt_bk[3:0] : BACK_COLOR_NUM; 					
			end
// delays
// Super Hires
		if (SHR  || LINEAR_SHR && ~TEXT) 
				begin
				hor_active_reg <= hor_active;						// register horizontal active signal
				hor_del2 <= hor_active_reg;						// delay 1 clock
				apple_video_addr_reg <= apple_video_addr;		// register the address
				end			
		else 
// text & low res graphics
      if(clk1 && allow_shift7 && ~hgr && ~SHR) 				// every clk1 for 80c, every other clk1 for 40c
			begin						
			if (count7 == 0) 											// only once/byte
				begin
				hor_active_reg <= hor_active;						// register horizontal active signal
				hor_del2 <= hor_active_reg; 						// delay 1 clock
				apple_video_addr_reg <= apple_video_addr;		// register the address		
				end
			end 
		else	
// high res graphics
      if(clk1 && hgr && ~SHR) 				
			begin
			if ((count7 == 0) && ~mode_80COL && h_count[0])	// once/byte - not 80 column mode
				begin
				hor_active_reg <= hor_active;						// register horizontal active signal
				hor_del2 <= hor_active_reg; 						// delay 1 clock
				apple_video_addr_reg <= apple_video_addr;		// register the address			
				end
			else
// double high res graphics & high res 80 columns
			if ((count7 == 0) && mode_80COL)						// once/byte - 80 column mode
				begin
				hor_active_reg <= hor_active;						// register horizontal active signal
				hor_del1 <= hor_active_reg; 						// delay 1 clock
				apple_video_addr_reg <= apple_video_addr;		// register the address
				end
			if (mode_80COL && (count7[0] == 1'b0)) 			// 80 columns mode & count7 is even 
				begin
				hor_del2 <= hor_del1;								// delay 1 clock 
				end
			end 
   end
//////////////////////////////////////////////////////////////////////////////////
// Super Hires graphics (SHR) data processing
// 640 mode, 320 mode and Fill (scanline interrupts not supported)
//
// scb_data is scanline control byte from BRAM via cb2.v
// scb_addr is scanline control byte address in BRAM sent to cb2.v to retrieve scb_data
// pal_data is 12 bit palette color from BRAM via cb2.v
// pal_addr is palette color address in BRAM sent to cb2.v to retrieve pal_data
// shr_color12 is the actual color for video output

	reg [7:0] shr_data = 0;							// local register to hold data from SRAM
	wire shr_resmode = scb_data[7]; 				// 1 if 640 wide resolution.  0 if 320 wide. 
// calculate when to use Fill 
	wire fill = (shr_resmode == 0) & (scb_data[5] == 1) & (((h_count[1] == 1'b0) && 
					(shr_data[7:4] == 0)) || ((h_count[1] == 1'b1) && (shr_data[3:0] == 0)));
	wire [3:0] palette_addr = scb_data[3:0];	// use lower 4 bits of scanline control byte for palette address
	wire [8:0] vfix = v_count - 48;  			
	wire [7:0] a2line = vfix[8:1]; 				// vfix/2
	wire [11:0] dith;
	wire do_dither;
	assign dith640 = shr_resmode & CBVID[0];
	reg [11:0] d0 = 12'h000;
	reg [11:0] d1 = 12'h000;

	always @(posedge clk50)
		begin
		if(SHR & clk1)											// when SHR is active and pixel clock is high
			begin	
			if(h_count[1:0] == 2'b11) shr_data <= ram_data[7:0];	// get data from SRAM via cb1.v every 4 pixels
			if(dith640)
				begin
				if(do_dither) shr_color12 <= dith;      // dither (dith is d1 mixed with d0)			
				else shr_color12 <= d1;						 // when not dithering
				end
			else	
			if (~fill) shr_color12 <= pal_data;			// if not Fill mode update the color from BRAM via cb1.v
			end													// after shifting to sync with dith
		end
// calculate address of scanline control byte in BRAM
	assign scb_addr = a2line;						
// calculate palette address in BRAM 	
	assign pal_addr = (shr_resmode == 1) ? ((h_count[1:0] == 2'b00) ? {palette_addr[3:0], 2'b10, shr_data[7:6]}:
							(h_count[1:0] == 2'b01) ? {palette_addr[3:0], 2'b11, shr_data[5:4]} :
							(h_count[1:0] == 2'b10) ? {palette_addr[3:0], 2'b00, shr_data[3:2]} :
							{palette_addr[3:0], 2'b01,shr_data[1:0]}):
							((h_count[1] == 1'b0) ? {palette_addr[3:0], shr_data[7:4]} : {palette_addr[3:0], shr_data[3:0]});

// dither - synthetic dither creating the 640 mode master colors
	always @(posedge clk1)
		begin
		d0 <= pal_data;		// puts the color from BRAM into a register (this will be the next pixel color)
		d1 <= d0;				// shift next color to the current pixel color
		end
	wire [4:0] r_dith = (d0[11:8] + d1[11:8]) >> 1;  	// average the red value of the 2 adjacent pixels
	wire [4:0] g_dith = (d0[7:4] + d1[7:4]) >> 1;		// average the green value of the 2 adjacent pixels
	wire [4:0] b_dith = (d0[3:0] + d1[3:0]) >> 1;		// average the blue value of the 2 adjacent pixels
	assign do_dither =  (h_count[0]) ? (d1 != d0) : 1'bz;	// although this implies an internal tri-state xst fixes it	
	assign dith = 	{ r_dith[3:0], g_dith[3:0], b_dith[3:0] };

//////////////////////////////////////////////////////////////////////////////////
// Double High Resolution graphics (DHR)
//////////////////////////////////////////////////////////////////////////////////
// these registers hold the 7 bit color data from SRAM
	reg [6:0] aux_data0 = 0;
	reg [6:0] main_data1 = 0;
	reg [6:0] aux_data2 = 0;
	reg [6:0] main_data3 = 0;
// time to get data from SRAM 
// (note these times make no sense but work)	

	wire ard0 = (count28 > 20) & (count28 < 26); 
	wire mrd1 = (count28 == 28) | (count28 < 5);  
	wire ard2 = (count28 > 6) & (count28 < 12); 
	wire mrd3 = (count28 > 13) & (count28 < 19);

	always @(posedge clk1)

		begin
		if (DHR) 
			begin
			if (ard0) aux_data0 <= ram_data[6:0];	// data from aux memory for 1st byte
			if (mrd1) main_data1 <= ram_data[6:0]; // data from main memory for 2nd byte
			if (ard2) aux_data2 <= ram_data[6:0];	// data from aux memory for 3rd byte
			if (mrd3) main_data3 <= ram_data[6:0];	// data from main memory for 4th byte
			end
		end 

	reg [7:0] bit_stream = 0;				// bits from main & aux memory combined
	wire new_color;							// used to find color change
	reg [3:0] barrel = 0;					// used to shift bits before being used as new color
					
// manipulate the bits to create a bit stream of 4 bit color index numbers
	always @(posedge clk50)
		begin
		if (clk1)
			begin
			case (count28[4:0]) 		
			5'b00000: bit_stream <= {main_data1[0], aux_data0[6:0]};
			5'b00100: bit_stream <= { main_data1[4:0], aux_data0[6:4] };
			5'b01000: bit_stream <= { aux_data2[1:0], main_data1[6:1] } ;
			5'b01100: bit_stream <= { aux_data2[5:0], main_data1[6:5] };
			5'b10000: bit_stream <= { main_data3[2:0], aux_data2[6:2] };
			5'b10100: bit_stream <= { main_data3[6:0], aux_data2[6] };
			5'b11000: bit_stream <= { aux_data0[3:0], main_data3[6:3] };
			5'b11100: ;				//	should never happen
			default:  bit_stream <= {bit_stream[0], bit_stream[7:1]};  	// rotate right so we use the next 4 bits
			endcase
			end
		end

	assign new_color = (count28[1] == 1'b0) || bit_stream[5:4] == bit_stream[1:0] ; 

// get the new color
	always @(posedge clk1)
		begin
		if (DHR & ~SHR & ~TEXT)
			begin
			if (new_color)
				begin
				dhrcolor_number <= {barrel[0], barrel[1], barrel[2], barrel[3]};
				end
			end
		end
// rotate the bits
	always @(*)
		begin
		case (count28[1:0])
		2'b00: barrel = bit_stream[3:0];
		2'b01: barrel = {bit_stream[2], bit_stream[1], bit_stream[0], bit_stream[3]};
		2'b10: barrel = {bit_stream[1], bit_stream[0], bit_stream[3], bit_stream[2]};
		2'b11: barrel = {bit_stream[0], bit_stream[3], bit_stream[2], bit_stream[1]};
		endcase
		end

//////////////////////////////////////////////////////////////////////////////////
// puts data from SRAM into shifter for hires & text. 
// LoRes data uses color table directly

	wire video_bit = shifter7[0];

	always @(posedge clk1 )
		begin
// data for high res graphics
		if( h_count[0] && hgr &&  ~DHR  && ~SHR )	
			begin
			vid_del <= video_bit;								// used by hires routines
			if (count7 == 6)			
				begin
				shifter7 <= ram_data[6:0];						// put SRAM data into shifter
				high_bit <= ram_data[7];						// used by hires routines
				end
			else 														// count7 does equal 6 so right shift
				begin
				shifter7 <= {1'b0, shifter7[7:1]};  		// right shift 1 bit
				end
			end
// data for text and low res graphics		
		else if(allow_shift7 && ~hgr  && ~SHR)  
				begin
					if(count7 == 0) 
						begin
						shifter7 <= (~ram_data[7]) ? char_data [7:0] : ~char_data[7:0];	// text data into shifter
						low_color_nibble <= v_count[3] ? ram_data[7:4] : ram_data[3:0];	// lores index number				
						flash  <= AltCharSet ? 0 : (ram_data[6] & ~ram_data[7]);				// enable/disable flashing text
						inverse <=  ~ram_data[7];														// enable/disable inverse
						end
					else												// not count7 == 0 so shift right						
						begin
						shifter7 <= {1'b0, shifter7[7:1]};  // right shift 1 bit
						end 	 
				end
		end 
// reverse lores index number bits for double lores
	assign dlr_nibble = {low_color_nibble[0], low_color_nibble[1],low_color_nibble[2],low_color_nibble[3]};
//////////////////////////////////////////////////////////////////////////////////
// determine next vertical character line
	wire spec  = allow_shift7 & (count7 == 0) & (hc == 64);
	wire next_vc = clk1 & spec & ((aux_byte == 0) | ~mode_80COL);
//////////////////////////////////////////////////////////////////////////////////
// high resolution graphics (HGR)
	wire odd_dot = h_count[1]; 
	reg [11:0] last12 = 12'h000;
	reg non_color = 0;							// non color flag
// hires colors per pixel 	
always @ (posedge clk1)
	begin
	last12 <= color12;
	if (video_bit && vid_del) 					// shifter7[0] and shifter7[0] delayed
			begin
			color12 <=12'hFFF;					// white
			if (shift_bitsum != 1) non_color <= 1'b1;
			end
	else	
	if(video_bit && ~vid_del)					// shifter7[0] and not shifter7[0] delayed
		begin
		if (non_color)
			begin 
			color12 <= last12;					// use last color again
			non_color <= 1'b0;					// reset non_color flag
			end
		else	
		if(odd_dot) 								// odd dot
			begin  
			if(high_bit) color12 <= 12'h49F;	// blue
			else color12 <= 12'hD2D;			// purple
			end 
		else											// even dot
			begin
			if(high_bit) color12 <= 12'hF60; // orange
			else color12 <= 12'h1D4;			// green
			end
		end
	else
	if (~video_bit && ~vid_del) 				// not shifter7[0] and not shifter7[0] delayed
		begin
		color12 <=12'h000;						// black  
		non_color <= 1'b1;						// set non color flag
		end		
	end	
////////////////////////////////////////////////////////////////////////////////// 
// horizontal character count (set & increment)
// aux_byte flag (set & toggled)
// non-SHR graphics delays
	always @(posedge clk50)
		begin
		if(clk1 && allow_shift7) 				// every clk1 for 80c, every other clk1 for 40c
			begin
			if(~(hor_active && vert_active)) // video display not active
				begin 
				hc <= 7'b0011000;					// start with 24 
				aux_byte <= GR ? 0 : 1;			// start with 0 for non-SHR graphics otherwise 1
				end    
			else 										// (hor_active && vert_active)
				begin 				
				if(count7 == 0)					
					begin
					aux_byte <= ~aux_byte; 		// flip aux_byte with each 7 bits
					GR_r <= GR;						// delay graphics 7 pixel clock cycles
					GR_r2 <= GR_r;					// delay graphics 7 more pixel clock cycles
					if((aux_byte == 1) || ~mode_80COL)  // always if not 80c othewise every other time
						begin												
						hc <= hc + 1;				// increment counter 24 to 63
						end 
					end  
				end 
			end 
		end
//////////////////////////////////////////////////////////////////////////////////
// vertical character count (set & incremented)
// flash count (incremented)
   always @(posedge clk50) 
		begin
		if (~vert_active) vc <= 5'b00000;			// vertical part of video display not active
		else 													// vertical part of video display active
			begin 
			if(next_vc) 
				begin 
				if((v_count == 48) || (vc == 24)) 	// end of top border or start of bottom border	
					begin
					flash_count <= flash_count +1; 	// increment flash count
					vc <= 5'b00000; 						// reset vertical character (line) count
					end   
				else  
					if (v_count[3:0] == 4'b1111)  vc <= vc + 1;  // increment char line every 16 scan lines
				end 
			end 
		end
////////////////////////////////////////////////////////////////////////////////// 
// decoding Apple soft switches 
	wire io_access = (bus_addr[15:8] == 8'b11000000);  //$C0xx area being accessed		
	reg init_settings = 1;

	always @(negedge Q3_s)
		begin
		if(ph1_s == 0)					// phase 1 low
			begin
			io_read <= 0;				// reset every time 
			if(reset_s == 1'b0) 
// reset
// set registers used by vga_video.v to default values
				begin
`ifdef tester
				test42 <= 8'b00000000;
`endif
				TEXT <= 1;
				PAGE2 <= 0;
				HIRES <= 0;
				MIXED <= 0;  
				AltCharSet <= 0;
				mode_80COL <= 0;
				mode_80STORE <= 0;
				DHR <= 0;
				SHR <= 0;
				WRITE_AUX <= 0;
				io_read <= 0;
				end 
// not reset
			else 
			if(io_access && (rw_s | ~rw_s)) 
// read or write in $C0xx area 
				begin
				case(bus_addr[7:0])
				8'h50:      TEXT <= 0;
				8'h51:      TEXT <= 1;
				8'h52:      MIXED <= 0;
				8'h53:      MIXED <= 1;
				8'h54:      PAGE2 <= 0;
				8'h55:      PAGE2 <= 1;   
				8'h56:      HIRES <= 0;
				8'h57:      HIRES <= 1;				  
				8'h5E:		  DHR <= 1;		// double hires on 
				8'h5F:		  DHR <= 0;		// double hires off	
				endcase 
// write softswitch					
				if(rw_s == 1'b0)
					begin           
					io_read <= 0;
					case(bus_addr[7:0])			
					8'h00: mode_80STORE <= 0; //$C000
					8'h01: mode_80STORE <= 1;
					8'h04:  WRITE_AUX <= 0;
					8'h05:  WRITE_AUX <= 1;
					8'h0C:  mode_80COL <= 0;
					8'h0D:  mode_80COL <= 1;				
					8'h0E:  AltCharSet <= 0;
					8'h0F:  AltCharSet <= 1; //$C00F
					8'h29: 
						begin 
						SHR <= bus_data[7];					// state of super hires on/off
						LINEAR_SHR <= bus_data[6];			// state of linear SHR mode
						monochrome <= bus_data[5];			// monochrome for DHR
						end
					8'h2A: CBVID <= bus_data[7:0];	// CB video bit 0 = 1 for SHR640 dither
																//				bit 1 = 1 for two grays
					8'h34: 									// $C034 lower 4 bits is color#
						begin
						border_flag <= 1'b1;
						BORDER_COLOR_NUM <= bus_data[3:0];
						end
`ifdef tester
					8'h42: test42 <= bus_data[7:0];	// input used for various tests
`endif

					8'h22:									// $C022 high 4 bits = text color# 
																//			low 4 bits = background color#
						begin
						tb_flag <= 1'b1;
						TEXT_COLOR_NUM <= bus_data[7:4];
						BACK_COLOR_NUM <= bus_data[3:0];
						end			
					endcase	// write
					end 		// write 
// read softswitch - puts data into created softswitches so Apple can read them 
				if(rw_s == 1'b1)	
					begin           
					case(bus_addr[7:0])				
					8'h22:									// $C022 high 4 bits = text color# 
																//			low 4 bits = background color#
						begin
						io_read <= 1;
						io_reg [7:0] <= ~tb_flag ? txt_bk : {TEXT_COLOR_NUM, BACK_COLOR_NUM};
						end
					8'h29: 
						begin 
						io_read <= 1;
						io_reg [7:0] <= {SHR,LINEAR_SHR,monochrome,5'b0};
						end
					8'h2A:
						begin 
						io_read <= 1;
						io_reg [7:0] <= CBVID;
						end						
					8'h34: 									// $C034 lower 4 bits is color#
						begin
						io_read <= 1;
						io_reg [3:0] <= ~border_flag ? bdr : {4'bz, BORDER_COLOR_NUM};
						end
					default: io_read <= 0;
					endcase	// read
					end		// read
				end 			// if io_access  
			end 				// if ph1 == 0
			else
				begin
// initial settings - runs once
				if (init_done & init_settings)
					begin
					CBVID <= cbvideo;			
					monochrome <= bw_dhr;
					init_settings <= 0;
					end
				end
		end

//////////////////////////////////////////////////////////////////////////////////
// sync apple clocks with main CB clock (from DCM)
   always @(posedge clk50) 
		begin
		rw_1 <= rw;
		rw_s <= rw_1;
		reset_1 <= reset_n;		// this probably shouldn't be done
		reset_s <= reset_1;		// it's sync'd in cb1.v
		end
//////////////////////////////////////////////////////////////////////////////////		
// changes border color to indicate reg or wire value
// bugborder is added to the border_color_num
		
`ifdef bbug
   always @(posedge clk50) 
		begin
		bugborder <= 1'b0;			// any single bit reg or wire can be placed here
		end
`endif
endmodule 
//////////////////////////////////////////////////////////////////////////////////		


